/************************************************************
* Copyrigth 2006 Miguel Angel Montano Medina UNAM           *
* Servidor web embebido                                     *
* Este software puede ser usado, copiado, modificado        *
* y distribuido por cualquier persona para usos personales  *
* o de investigacion siempre y cuando se mantenga esta      *
* licencia presente. No se autoriza el uso de este software *
* para fines comerciales.                                   *
************************************************************/

#ifndef _TCP_C_
#define _TCP_C_

//#define RTX_DEBUG
//#define TCP_DEBUG

#include "uinet/ethernet.h"
#include "uinet/tcp.h"
#include "uinet/ip.h"

#define  set_packet32(d,s) packet[d] = make8(s,3); \
                           packet[d+1] = make8(s,2); \
                           packet[d+2] = make8(s,1); \
                           packet[d+3]= make8(s,0);


void tcp_env_reset(){
   uint32 seq;
   uint32 ack;
   uint16 chksum16,t16;
   uint temp;

   if(packet[TCP_hdrflags+1] & ACK){
      seq=make32(packet[TCP_acknum],packet[TCP_acknum+1],packet[TCP_acknum+2],packet[TCP_acknum+3]);
      ack=0;
      packet[TCP_hdrflags+1] = RST;
   }
   else{
      t16=packet[TCP_hdrflags] >> 4;
      t16+=(packet[ip_vers_len] & 0x0F);
      t16*=4;
      t16=make16(packet[ip_pktlen],packet[ip_pktlen+1])-t16;
      seq=0;
      ack=make32(packet[TCP_seqnum],packet[TCP_seqnum+1],packet[TCP_seqnum+2],packet[TCP_seqnum+3])+t16;
      if(packet[TCP_hdrflags+1] & SYN){ // ||(packet[TCP_hdrflags+1] & FIN))?
         ack++;
      }
      packet[TCP_hdrflags+1] = RST|ACK;
   }

   temp = packet[TCP_destport];
   packet[TCP_destport]   = packet[TCP_srcport];
   packet[TCP_srcport]    = temp;

   temp = packet[TCP_destport+1];
   packet[TCP_destport+1] = packet[TCP_srcport+1];
   packet[TCP_srcport+1]  = temp;

   packet[TCP_hdrflags]=0x50;
   packet[TCP_window]=0x0;
   packet[TCP_window+1]=0x0;

   set_packet32(TCP_seqnum,seq);
   set_packet32(TCP_acknum,ack);

   packet[TCP_cksum] = 0x00;
   packet[TCP_cksum+1] = 0x00;

   packet[PSEUDO_ZERO]=0x00;
   packet[PSEUDO_PTCL]=PROT_TCP;
   packet[PSEUDO_LEN]=0;
   packet[PSEUDO_LEN+1]=20;
   packet[ip_destaddr]=packet[ip_srcaddr];
   packet[ip_destaddr+1]=packet[ip_srcaddr+1];
   packet[ip_destaddr+2]=packet[ip_srcaddr+2];
   packet[ip_destaddr+3]=packet[ip_srcaddr+3];
   packet[ip_srcaddr]=MYIP[0];
   packet[ip_srcaddr+1]=MYIP[1];
   packet[ip_srcaddr+2]=MYIP[2];
   packet[ip_srcaddr+3]=MYIP[3];

   chksum16=cksum(&packet[PSEUDO_ZERO],32);

   packet[TCP_cksum+1] = make8(chksum16,0);
   packet[TCP_cksum] = make8(chksum16,1);

   ip(PROT_TCP, 20, 0x80, &packet[ip_destaddr], &packet[enetpacketSrc0]);
}

void tcp_re_enviar(uint flags, uint16 num_datos){
   cnte->sndnxt -= num_datos;
   if(flags&SYN || flags&FIN){
      cnte->sndnxt--;
   }
   tcp_enviar(flags,num_datos);
   if(flags&SYN || flags&FIN){
      cnte->sndnxt++;
   }
}

void tcp_enviar(uint flags, uint16 num_datos){
   uint16 chksum16;

   packet[TCP_srcport]    = MY_PORT_ADDRESS_H;
   packet[TCP_srcport+1]  = MY_PORT_ADDRESS_L;

   packet[TCP_destport]   = cnte->pto[0];
   packet[TCP_destport+1] = cnte->pto[1];

   set_packet32(TCP_seqnum,cnte->sndnxt);
   set_packet32(TCP_acknum,cnte->rcvnxt);

   if(num_datos!=0){
      cnte->lude = num_datos;
      cnte->sndnxt += num_datos;
      cnte->tempo=0;           // iniciar el contador de tiempo
   }

   if(flags & SYN){ // enviar options
      packet[TCP_hdrflags] = 0x60;
      packet[TCP_data]=0x02;    // MSS kind
      packet[TCP_data+1]=0x04;  // MSS length
      packet[TCP_data+2]=cnte->MSS>>8;  // MSS max size
      packet[TCP_data+3]=cnte->MSS;  //
      num_datos=num_datos+24;
      cnte->tempo=0;           // iniciar el contador de tiempo
   }else{
      packet[TCP_hdrflags] = 0x50;
      num_datos=num_datos+20;
   }

   packet[TCP_hdrflags+1] = flags;

   packet[TCP_window]=cnte->MSS>>8;  // MSS max size
   packet[TCP_window+1]=cnte->MSS;  //


   packet[TCP_cksum] = 0x00;
   packet[TCP_cksum+1] = 0x00;

   packet[PSEUDO_ZERO]=0x00;
   packet[PSEUDO_PTCL]=PROT_TCP;
   packet[PSEUDO_LEN]=make8(num_datos,1);
   packet[PSEUDO_LEN+1]=make8(num_datos,0);
   packet[ip_destaddr]=cnte->ip[0];
   packet[ip_destaddr+1]=cnte->ip[1];
   packet[ip_destaddr+2]=cnte->ip[2];
   packet[ip_destaddr+3]=cnte->ip[3];
   packet[ip_srcaddr]=MYIP[0];
   packet[ip_srcaddr+1]=MYIP[1];
   packet[ip_srcaddr+2]=MYIP[2];
   packet[ip_srcaddr+3]=MYIP[3];

   chksum16=cksum(&packet[PSEUDO_ZERO],num_datos+12);

   packet[TCP_cksum+1] = make8(chksum16,0);
   packet[TCP_cksum] = make8(chksum16,1);

   ip(PROT_TCP, num_datos, 0x80, &cnte->ip[0], &cnte->mac[0]);
/*
   packet[ip_vers_len]=0x45;
   packet[ip_tos]=0;

   num_datos = num_datos + 20;

   packet[ip_pktlen]=make8(num_datos,1);
   packet[ip_pktlen+1]=make8(num_datos,0);

   packet[ip_id]=make8(ip_id_num,1);
   packet[ip_id+1]=make8(ip_id_num,0);
   ip_id_num++;

   packet[ip_frag_offset]=0x40;
   packet[ip_frag_offset+1]=0;

   packet[ip_ttl]=0x80;
   packet[ip_proto]=PROT_TCP;

   packet[ip_hdr_cksum]=0x00;
   packet[ip_hdr_cksum+1]=0x00;

   chksum16=cksum(&packet[ip_vers_len],20);

   packet[ip_hdr_cksum] = make8(chksum16,1);
   packet[ip_hdr_cksum+1] = make8(chksum16,0);


   packet[enetpacketDest0]=cnte->mac[0];
   packet[enetpacketDest1]=cnte->mac[1];
   packet[enetpacketDest2]=cnte->mac[2];
   packet[enetpacketDest3]=cnte->mac[3];
   packet[enetpacketDest4]=cnte->mac[4];
   packet[enetpacketDest5]=cnte->mac[5];

   packet[enetpacketSrc0]=MAC_ADDR_0;
   packet[enetpacketSrc1]=MAC_ADDR_1;
   packet[enetpacketSrc2]=MAC_ADDR_2;
   packet[enetpacketSrc3]=MAC_ADDR_3;
   packet[enetpacketSrc4]=MAC_ADDR_4;
   packet[enetpacketSrc5]=MAC_ADDR_5;

   packet[enetpacketType0]=0x8;
   packet[enetpacketType1]=0x0;

   if(flags & SYN){
      nic_putd(packet,58);
   }else{
      nic_putd(packet,num_datos+14);
   }*/
}

void tcp_init(){
   uint i;
   uint s0,s1,s2,s3;
   //ini_seqnum = 0x30011978;

   i2c_start();
   i2c_write(DS1338_I2C_WRITE_ADDR);
   i2c_write(DS1338_INISEQ_0);
   i2c_start();
   i2c_write(DS1338_I2C_READ_ADDR);
   s0=i2c_read();
   s1=i2c_read();
   s2=i2c_read();
   s3=i2c_read(0);
   i2c_stop();
// printf("\r\n%u %u %u %u",s0,s1,s2,s3);
   ini_seqnum=make32(s0,s1,s2,s3);

   for(i=0,cnte=clientes;i<MAX_CLIENTES;cnte++,i++){
      cnte->id=i;
      cnte->estado=0;
   }
   tcp_activas=0;
   printf("\r\nMSS: %lu",MAX_MSS);
   printf("\r\nINET: %d sockets listos",MAX_CLIENTES);
}


sint tcp_conectar(){
   uint i;
   tcp_activas=0;
   if(!(packet[TCP_destport] == MY_PORT_ADDRESS_H && packet[TCP_destport+1] == MY_PORT_ADDRESS_L)){
      return -1;
   }

   for(i=0,cnte=clientes;i<MAX_CLIENTES;i++,cnte++){
      if(cnte->estado!=0){
         tcp_activas++;
         if(packet[ip_srcaddr]      == cnte->ip[0]
            && packet[ip_srcaddr+1] == cnte->ip[1]
            && packet[ip_srcaddr+2] == cnte->ip[2]
            && packet[ip_srcaddr+3] == cnte->ip[3]
            && packet[TCP_srcport]  == cnte->pto[0]
            && packet[TCP_srcport+1]== cnte->pto[1]){
            return i; /* si estaba hablando contigo, te atiendo */
          }
      }
   }
   if(packet[TCP_hdrflags+1] & SYN){
      for(i=0,cnte=clientes;i<MAX_CLIENTES;i++,cnte++){
         if(cnte->estado==0){
            cnte->estado=1;
            cnte->mac[0]=packet[enetpacketSrc0];
            cnte->mac[1]=packet[enetpacketSrc0+1];
            cnte->mac[2]=packet[enetpacketSrc0+2];
            cnte->mac[3]=packet[enetpacketSrc0+3];
            cnte->mac[4]=packet[enetpacketSrc0+4];
            cnte->mac[5]=packet[enetpacketSrc0+5];
            cnte->ip[0]=packet[ip_srcaddr];
            cnte->ip[1]=packet[ip_srcaddr+1];
            cnte->ip[2]=packet[ip_srcaddr+2];
            cnte->ip[3]=packet[ip_srcaddr+3];
            cnte->pto[0]=packet[TCP_srcport];
            cnte->pto[1]=packet[TCP_srcport+1];
            cnte->sndnxt=++ini_seqnum;

            i2c_start();
            i2c_write(DS1338_I2C_WRITE_ADDR);
            i2c_write(DS1338_INISEQ_0);
            i2c_write(make8(ini_seqnum,3));
            i2c_write(make8(ini_seqnum,2));
            i2c_write(make8(ini_seqnum,1));
            i2c_write(make8(ini_seqnum,0));
            i2c_stop();

            cnte->rcvnxt=make32(packet[TCP_seqnum],packet[TCP_seqnum+1],packet[TCP_seqnum+2],packet[TCP_seqnum+3]);
            cnte->lude=0;
            if(packet[TCP_data]==2&&packet[TCP_data+1]==4){
               cnte->MSS=make16(packet[TCP_data+2],packet[TCP_data+3]);
               if(cnte->MSS>MAX_MSS){
                  cnte->MSS=MAX_MSS;
               }
            }
            else{
               cnte->MSS=536;
            }
            cnte->tempo=0;   // el tiempo en el que emiti el ultimo paquete con datos / tiempo para retransmitir
            cnte->rto=30;    // prueba en 3 segundos
            cnte->sa=0;      // estimadores
            cnte->sv=30;      // estimadores
            cnte->nrtx=0;    // cero retransmisiones
            return i; /* bienvenido */
         }
      }
   }
   return -1;  /* picatelo!! */
}




void tcp_cerrar(uint id){
   cnte=&clientes[id];
   if(cnte->estado==3 && cnte->nrtx==0){   // si estoy en established y no hay retransmisiones en proceso
      tcp_enviar(FIN|ACK,0);           // procedo al cierre normal.
      cnte->lude=1;
      cnte->sndnxt++;
      cnte->estado=5;
   }
   else{                            // si estoy haciendo una retransmision no espero, aborto y cierro la conexion
      cnte->estado=0;
   }
}



void tcp(){
	uint flags;
   uint16 tmp;
   sint temp;
   uint32 seg;
   uint32 t32;
   uint16 seglen;
   uint acked;

   flags=packet[TCP_hdrflags+1];

   temp=tcp_conectar();

   if(temp==-1){
      if(flags & RST){   // si el paquete trae un reset, ignorarlo
#ifdef TCP_DEBUG
         printf("\r\n%d return 1",cnte->id);
#endif
         return;
      }
      tcp_env_reset();  // si no enviar un reset
#ifdef TCP_DEBUG
      printf("\r\n%d return 2",cnte->id);
#endif
      return;
   }
   cnte=&clientes[temp];

   if(cnte->estado!=1){
      temp=packet[TCP_hdrflags] >> 4;
      temp+=(packet[ip_vers_len] & 0x0F);
      temp*=4;
      seglen=make16(packet[ip_pktlen],packet[ip_pktlen+1])-temp;
      seg=make32(packet[TCP_seqnum],packet[TCP_seqnum+1],packet[TCP_seqnum+2],packet[TCP_seqnum+3]);

      if((seg<cnte->rcvnxt)||(seg>=cnte->rcvnxt+cnte->MSS)){
         if(flags & RST){
#ifdef TCP_DEBUG
            printf("\r\n%d return 3",cnte->id);
#endif
            return;
         }
         tcp_enviar(ACK,0);
#ifdef TCP_DEBUG
         printf("\r\n%d return 4",cnte->id);
#endif

         return;
      }
      if(seglen>0){
         t32=seg+seglen-1;
         if((t32<cnte->rcvnxt)||(t32>=cnte->rcvnxt+cnte->MSS)){
            if(flags & RST){
#ifdef TCP_DEBUG
               printf("\r\n%d return 5",cnte->id);
#endif
               return;
            }
            tcp_enviar(ACK,0);
#ifdef TCP_DEBUG
            printf("\r\n%d return 6",cnte->id);
#endif

            return;
         }
      }
      if(flags & RST){
         cnte->estado=0;
         ev_tcp_fin(); // notificar a la aplicacion del cierre de conexion (abortar)
#ifdef TCP_DEBUG
         printf("\r\n%d return 7",cnte->id);
#endif

         return;
      }
      if(flags & SYN){
         tcp_env_reset();
         cnte->estado=0;
         ev_tcp_fin(); // notificar a la aplicacion del cierre de conexion (abortar)
#ifdef TCP_DEBUG
         printf("\r\n%d return 8",cnte->id);
#endif

         return;
      }
      if(flags & ACK){
         seg=make32(packet[TCP_acknum],packet[TCP_acknum+1],packet[TCP_acknum+2],packet[TCP_acknum+3]);
         // si el ack es un duplicado
         if(seg<(cnte->sndnxt-cnte->lude)) {
#ifdef TCP_DEBUG
               printf("\r\n%d return 9",cnte->id);
#endif
            return;           // ignorarlo
         }
         // si el ack es de algo que no he mandado aun, envio un ack y lo ignoro.
         // o si el ack es solo de una parte de lo que envie(muy raro), mando un ack y espero el paquete completo.
         // DUDA DUDA
         else if(seg>cnte->sndnxt){
            tcp_enviar(ACK,0);
#ifdef TCP_DEBUG
            printf("\r\n%d return 10",cnte->id);
#endif
            return;
         }else{   //ack valido!
//          cnte->sndnxt=seg;
            if(cnte->lude){
               if(cnte->nrtx==0){ // no tomar muestra de una retransmision [KARN]
                  sint m;        // Algoritmo de Van Jaconson
                  m=cnte->tempo; // tempo es el RTT, ya que inicio en 0 al emitir el ultimo segmento.
                  m-= (cnte->sa >> 3);
                  cnte->sa += m;          // nuevo valor de sa
                  if (m < 0)
                     m = -m;
                  m -= (cnte->sv >> 2);
                  cnte->sv += m;          // nuevo valor de sv
                  cnte->rto = (cnte->sa >> 3) + cnte->sv; // nuevo rto
#ifdef RTX_DEBUG
                  printf("\r\n rtt:%u rto:%u sa:%u, sv:%u",cnte->tempo,cnte->rto,cnte->sa,cnte->sv);
#endif
               }
               else{ // llego el ack de una retransmision, iniciar en cero
                  cnte->nrtx=0; // el rto se mantiene para el siguiente paquete [KARN]
#ifdef RTX_DEBUG
                  printf("\r\n despues de rtx rtt:%u rto:%u sa:%u, sv:%u",cnte->tempo,cnte->rto,cnte->sa,cnte->sv);
#endif
               }
            }
            cnte->lude=0; // ya no tengo datos pendientes
            cnte->tempo=0;
            acked=1;
         }
      }
      else{  // si no esta el flag ACK ignorar el paquete
#ifdef TCP_DEBUG
         printf("\r\n%d return 11",cnte->id);
#endif
         return;
      }
   }

   switch(cnte->estado){
   case 1: /* Listen 1*/
      if(flags & RST){ //reset ignorado
#ifdef TCP_DEBUG
         printf("\r\n%d return 12",cnte->id);
#endif

         return;
      }
      if(flags & ACK){ // error no puede llegar un ack aqui
         tcp_env_reset();
#ifdef TCP_DEBUG
         printf("\r\n%d return 13",cnte->id);
#endif
         return;
      }
#ifdef TCP_DEBUG
      printf("\r\n%u MSS %lu",cnte->id,cnte->MSS);
#endif
      cnte->rcvnxt++;
      tcp_enviar(SYN|ACK,0);
      cnte->lude=1;
      cnte->sndnxt++;

      cnte->estado=2;
#ifdef TCP_DEBUG
         printf("\r\n%d estado 1",cnte->id);
#endif
      break;
   case 2:  /*syn_rcvd*/
      if(acked==1){
         cnte->estado=3;
         cnte->lude=0;
         ev_tcp_ini();
#ifdef TCP_DEBUG
         printf("\r\n%d estado 2",cnte->id);
#endif
      }
      else{
#ifdef TCP_DEBUG
         printf("\r\n%d estado 2 no ack",cnte->id);
#endif
         }

      break;
   case 3:  /*established*/
      switch(flags){
      case  (PSH|ACK):
         cnte->rcvnxt+=seglen;
         // enviar ack
//       tcp_enviar(ACK,0);
         // procesar
         temp=ev_tcp_psh();
#ifdef TCP_DEBUG
         printf("\r\n%d estado 3 pshack",cnte->id);
#endif

         if(temp==-1){
#ifdef TCP_DEBUG
         printf("\r\n%destado 3 cierra",cnte->id);
#endif
            tcp_enviar(FIN|ACK,0);
            cnte->lude=1;
            cnte->sndnxt++;
            cnte->estado=5;
         }
         break;
      case (FIN|ACK):
         cnte->rcvnxt++;
         tcp_enviar(ACK,0);
#ifdef TCP_DEBUG
         printf("\r\n%d estado 3 finack",cnte->id);
#endif

//       fin_aplicacion(); // notificar a la aplicacion que ya nos cortaron!
         tcp_enviar(FIN|ACK,0);
         cnte->lude=1;
         cnte->sndnxt++;
         cnte->estado=4;
         break;
      case ACK:
         if(acked==1){
#ifdef TCP_DEBUG
         printf("\r\n%d estado 3 ack",cnte->id);
#endif

            temp=ev_tcp_ack();
            if(temp==-1){
#ifdef TCP_DEBUG
               printf("\r\n%d estado 3 enviando finack",cnte->id);
#endif
               tcp_enviar(FIN|ACK,0);
               cnte->lude=1;
               cnte->sndnxt++;
               cnte->estado=5;
            }
         }
         else{
#ifdef TCP_DEBUG
         printf("\r\n%d estado 3 ack no ack",cnte->id);
#endif
         }
         break;
      default:
#ifdef TCP_DEBUG
         printf("\r\n%d estado 3 default",cnte->id);
#endif
         break;
      }
      break;
   case 4: /* close_wait last_ack */
      if(acked==1){
#ifdef TCP_DEBUG
         printf("\r\n%d estado 4",cnte->id);
#endif
         cnte->lude=0;
         cnte->estado=0;
         ev_tcp_fin(); // notificar a la aplicacion que ya nos cortaron! (0)
      }
      else{
#ifdef TCP_DEBUG
         printf("\r\n%d estado 4 no ack",cnte->id);
#endif
         }

      break;
   case 5: /* fin_wait_1 */
      switch(flags){
      case (FIN|ACK):
#ifdef TCP_DEBUG
         printf("\r\n%d estado 5 fin|ack",cnte->id);
#endif
         cnte->rcvnxt++;
         tcp_enviar(ACK,0);
         cnte->lude=0;
         if(acked==1){
            cnte->estado=8;
            cnte->tempo=0;
            ev_tcp_fin(); // notificar a la aplicacion del cierre de conexion
         }
         else
            cnte->estado=7;
         break;
      case ACK:
#ifdef TCP_DEBUG
         printf("\r\n%d estado 5 ack",cnte->id);
#endif
         cnte->lude=0;
         cnte->estado=6;
         break;
      default:
#ifdef TCP_DEBUG
         printf("\r\n%d estado 5 default",cnte->id);
#endif
      }
      break;
   case 6: /* fin_wait_2 */
      if(flags==FIN|ACK){
#ifdef TCP_DEBUG
         printf("\r\n%d estado 6 fin",cnte->id);
#endif
         cnte->rcvnxt++;
         tcp_enviar(ACK,0);
         cnte->lude=0;
         cnte->estado=8;
         cnte->tempo=0;
         ev_tcp_fin(); // notificar a la aplicacion del cierre de conexion exitoso
      }
      else{
#ifdef TCP_DEBUG
         printf("\r\n%d estado 6 otro flags",cnte->id);
#endif
      }
      break;
   case 7: /* closing */
      if(flags==ACK){
#ifdef TCP_DEBUG
         printf("\r\n%d estado 7 ack",cnte->id);
#endif
         cnte->lude=0;
         cnte->estado=8;
         cnte->tempo=0;
         ev_tcp_fin(); // notificar a la aplicacion del cierre de conexion exitoso
      }
      else{
#ifdef TCP_DEBUG
         printf("\r\n%d estado 7 otro flags",cnte->id);
#endif
      }

      break;
   case 8: /* time wait */
#ifdef TCP_DEBUG
      printf("\r\n%d estado 8",cnte->id);
#endif
      break;
   }
}


void tcp_rtx(){
   uint i;
   for(i=0,cnte=clientes;i<MAX_CLIENTES;cnte++,i++){
      if(cnte->estado!=0){
         if(cnte->estado==8){
            cnte->tempo++;
            if(cnte->tempo==0){  // 2MSL=25.6 segundos;
#ifdef RTX_DEBUG
               printf("\r\nCerrado! time wait %d",i);
#endif
               ev_tcp_fin();
               cnte->estado=0;
            }
            continue;
         }

         if(cnte->lude){
            cnte->tempo++;
            if(cnte->tempo==cnte->rto){ // retransmitir
#ifdef RTX_DEBUG
               cnte->nrtx++; // incremento el contador de retransmisiones
               printf("\r\n %d rtx rto:%u nrtx:%u",cnte->id, cnte->rto,cnte->nrtx);
               printf("\r\n %d retransmitir!",cnte->id);
#endif
               if(cnte->nrtx==7){   // 7 retransmisiones y nada
            		cnte->estado=0;   // adios!
#ifdef RTX_DEBUG
            		printf("\r\ncliente %d libre por num max de retx",i);
#endif
            		continue;
               }

               cnte->rto<<=1;       //  exponential back off
               if(cnte->rto>127)
                  cnte->rto=255;       //  limite del exp back off

               switch(cnte->estado){
               case 1: // Listen 1
#ifdef RTX_DEBUG
                  printf("\r\ntx estado 1?");
#endif
                  break;
               case 2:  // syn_rcvd
#ifdef RTX_DEBUG
                  printf("\r\nrtx synack");
#endif
                  tcp_re_enviar(SYN|ACK,0);
                  break;
               case 3:  // established
                  // procesar
#ifdef RTX_DEBUG
                  printf("\r\nrtx app");
#endif
                  ev_tcp_rtx();
                  break;
               case 4: // close_wait last_ack
               case 5: // fin_wait_1
#ifdef RTX_DEBUG
                  printf("\r\nrtx finack");
#endif
                  tcp_re_enviar(FIN|ACK,0);
                  break;
               }
            }
         }
      }
   }
}


#endif

