/************************************************************
* Copyrigth 2006 Miguel Angel Montano Medina UNAM           *
* Servidor web embebido                                     *
* Este software puede ser usado, copiado, modificado        *
* y distribuido por cualquier persona para usos personales  *
* o de investigacion siempre y cuando se mantenga esta      *
* licencia presente. No se autoriza el uso de este software *
* para fines comerciales.                                   *
************************************************************/

/************************************************************
* Archivo principal. Para compilar: hacer este archivo 
* el archivo de proyecto.
************************************************************/

//#include <18F452.h>
#include <18F4525.h>
#use delay(clock=40000000)
#fuses HS,H4,NOWRT,NOWDT,WDT256,PUT,NOPROTECT,NOBROWNOUT,NOLVP,NOCPD,NOEBTR
#use rs232(baud=115200,xmit=PIN_C6, rcv=PIN_C7)
#use i2c(MASTER, SDA=PIN_C4, SCL=PIN_C3, FORCE_HW,FAST)
#id 0x0812
#case

#define VMAY 1
#define VMEN 0

//#define debug
#include <stdlib.h>
#include "arch/tipos.h"
#include "uinet/globales.h"
#include "devs/realtek.c"
#include "devs/fs.c"
#include "devs/ds1338.c"
#include "uinet/uinet.h"

#include "xmodem/xmodem-eeprom.c"
#include "mconfig.c"
#include "http.c"



void multicast(){
   uint i;
   char *udp_buf;
   udp_buf=&packet[UDP_data];

   i2c_start();
   i2c_write(0xC1);
   for(i=0;i<31;i++,udp_buf++){
      *udp_buf=i2c_read();
   }
   *udp_buf=i2c_read(0);
   i2c_stop();

   udp_buf=&packet[UDP_data];

   udp_enviar(32,MULTICAST);
   if(usr_udp)
      udp_enviar(32,UNICAST);
}


void get_packet()
{
   uint lcip;
   uint16 ldat;
   uint offip;
   if(es_arp){
      if(es_peticion_arp ){
         output_bit(PIN_A5,1);
         arp_env_mac();
         output_bit(PIN_A5,0);
      }
      if(es_respuesta_arp){
         int i;
         for(i=0;i<6;i++)
            t_mac[i]=packet[arp_shaddr+i];
         arp_resp=1;
      }

   }
   if(es_ip){
      if(es_mi_ip||es_all_systems||es_mi_grp){
         lcip=(packet[ip_vers_len]&0x0F)<<2;                // logitud de la cabecera IP
         if(cksum(&packet[ip_vers_len],lcip)==0x0000){      // checksum IP ok ?
            offip=(packet[ip_vers_len]-0x45)*4;             // offset al inicio de datos
            ldat=make16(packet[ip_pktlen],packet[ip_pktlen+1])-lcip;
            switch(packet[ip_proto]){
            case PROT_ICMP:
               output_bit(PIN_A5,1);
               if(cksum(&packet[ICMP_type+offip],ldat)==0x0000){  // checksum ICMP ok ?
               output_bit(PIN_A5,0);
               delay_us(50);
               output_bit(PIN_A5,1);
               icmp();
               output_bit(PIN_A5,0);
               }
               break;
            case PROT_TCP:
               tcp();
               break;
            case PROT_UDP:
               udp();
               break;
            case PROT_IGMP:
               switch(packet[IGMP_ver_type+offip]){
               case 0x11:                             // es un query.
                     IGMP_REPORT=1;
                     igmp_fuera=c_tiempo+(packet[IGMP_mrt+offip]>>2); // en 1/4 del tiempo max, contesto (c_tiempo e IGMP_mrt tienen las mismas unidades 1/10[s])
                  break;
               case 0x16:
               case 0x12:                             // es un reporte v1 o v2 de mi grupo
                     IGMP_REPORT=0;                   // ok, ya no me reporto
                  break;
               }
               break;
            }
         }
      }
   }
}


void ini_pic(){
   setup_adc_ports(NO_ANALOGS);
   setup_adc(ADC_OFF);
   output_float(PIN_C4);
   output_float(PIN_C3);

   set_tris_a(0xFF);
   setup_psp(PSP_DISABLED);
   setup_wdt(WDT_OFF);
   setup_timer_1(T1_DISABLED);
   setup_timer_2(T2_DISABLED,0,1);
   setup_timer_3(T3_DISABLED);
}


void control_init(){
   printf("\r\nIniciando controlador red inalambrica...");
   // reset al controlador
   output_bit(PIN_A3,0);
   delay_ms(1);
   output_bit(PIN_A3,1);
}


void main()
{
   uint tcp_fuera, http_fuera;
   ini_pic();
   switch ( restart_cause() )
   {
      case WDT_TIMEOUT:
      {
         printf("\r\nWatchdog timeout!");
         break;
      }
      case NORMAL_POWER_UP:
      {
         printf("\r\nPower up!");
         break;
      }
   }
   setup_wdt(WDT_ON);
   output_bit(PIN_A4,1);   // LED encendido
   output_bit(PIN_A5,0);
   nic_init();
   printf("\r\n\n******************************************************");
   printf("\r\nServidor petit v%u.%u",VMAY,VMEN);
   printf("\r\npetit: iniciando...");
   setup_wdt(WDT_OFF);
   timers_init();
   obtener_configuracion();
   
   if(arp_init()==1)
      guardar_configuracion();
            
   setup_wdt(WDT_ON);
   tcp_init();
   http_init();
   control_init();
   printf("\r\npetit: listo!");
   igmp_report();
   usr_udp=0;
   tcp_fuera=c_tiempo+1;
   http_fuera=c_tiempo+10;
   igmp_fuera=c_tiempo+10; // en un segundo va de nuevo el reporte igmp
   udp_fuera=c_tiempo+50; // 5 sec max de espera para 'cliente presente'
   IGMP_REPORT=1;
   ARP_PET=0;
   while(1){
      restart_wdt();
      packet_len=nic_begin_packet_rx();

      if(packet_len>0){
         nic_packet_get_data(packet,packet_len);
         nic_end_packet_rx();
         packet[packet_len]='\0';
         get_packet();
      }

      if((arp_fuera==c_tiempo) && ARP_PET==1){
         ARP_PET=0;
         http_app2(arp_resp);
      }

      if(tcp_fuera==c_tiempo){ // cada 100 ms
         tcp_fuera=c_tiempo+1;
         tcp_rtx();
      }

      if(http_fuera==c_tiempo){ // cada segundo
         http_fuera=c_tiempo+10;
         http_tfuera();
      }

      if((igmp_fuera==c_tiempo) && IGMP_REPORT==1){
         IGMP_REPORT=0;
         igmp_report();
      }

      if((udp_fuera==c_tiempo) && usr_udp==1){
         usr_udp=0;
      }

      if(input(PIN_A2)){
         multicast();
      }

   }
}

