/************************************************************
* Copyrigth 2006 Miguel Angel Montano Medina UNAM           *
* Servidor web embebido                                     *
* Este software puede ser usado, copiado, modificado        *
* y distribuido por cualquier persona para usos personales  *
* o de investigacion siempre y cuando se mantenga esta      *
* licencia presente. No se autoriza el uso de este software *
* para fines comerciales.                                   *
************************************************************/

#include <16F819.h>
#fuses HS,WDT,NOPROTECT,NOLVP,MCLR,PUT
#use delay(clock=20000000)
#use rs232(baud=57600, xmit=PIN_A0, rcv=PIN_A1)  // Jumpers: 8 to 11, 7 to 12
#use i2c(SLAVE,SDA=PIN_B1,SCL=PIN_B4,ADDRESS=0xC0,FORCE_HW,FAST)

#case

#define CONTROL


#byte OSCCON=0x8f
#bit  IOFS = 0x8F.2


#define MIDIR  0xA0

#include "nrf2401.c"
#include "defs.h"

int contador_t1;
int id_paq;

int1 nuevo,rw;

int nsensores;       // numero de sensores detectados
int16 sensores;      // cada bit representa 1 sensor, 1: activo, 0: inactivo
int16 dlistos;       // representa que sensor ya transmitio sus datos;
union{
   int16 l[16];  // representa los datos que cada sensor envio
   int   b[32];
} muestras;

int index;


#byte SSPSTAT = 0x94

int i2c_estado;
int dummy,registro;
typedef enum {ESPERANDO, REGISTRO, ACCION} estados;



#int_timer1
void beacon(){
   int i;
   set_timer1(3036);
   contador_t1++;
   if(contador_t1==10){
      id_paq++;
      sensores=sensores&dlistos;
      dlistos=0;
      output_bit(PIN_A3,1);
      contador_t1=0;
      paquete[DST]= BCDIR;
      paquete[RMT]= MIDIR;
      paquete[SEQ]=id_paq;
      paquete[FLAGS]=0x00;
      paquete[ID] =BEACON;
      set_tx_mode();
      write_nrf(paquete);
      output_bit(PIN_A3,0);
      set_rx_mode();
      for(i=0;i<32;i=i+2){
         muestras.b[i]=200;
         muestras.b[i+1]=0xCE; // -50
      }  
   }
}


#int_ext
void int_externa(void){
   int16 s;
   int  n;
   int1 found;
   read_nrf(paquete);
   if(paquete[DST] == MIDIR){
      if(paquete[ID] == TXSINC){
         output_bit(PIN_A2,1);
         n=paquete[RMT]&0x0f;    // obtener la direccion del sensor
         s=1<<n;                 // crear una mascara con el bit del sensor, ej: sensor3 -> 1000
         if(sensores & s){       // verificar si el sensor ya habia transmitido antes
            paquete[FLAGS]=ACK;  // enviar un ack
         }
         else{                   // no habia transmitido antes
            sensores=sensores|s; // "registrarlo"
            paquete[FLAGS]=SYNC|ACK; // forzar su resincronizacion;
         }

         dlistos=dlistos|s;      // este sensor ya cumplio
         muestras.l[n]=make16(paquete[DATOS],paquete[DATOS+1]); // guardar la muestra
         paquete[DST]=paquete[RMT];
         paquete[RMT]=MIDIR;
         paquete[SEQ]=paquete[SEQ];
         paquete[ID]=TXSINC;
         set_tx_mode();
         write_nrf(paquete);
         set_rx_mode();
         if(dlistos==sensores){
            nuevo=1;
         }
         output_bit(PIN_A2,0);
      }
   }
}

#INT_SSP
void ssp_interupt (){
   int s;
   s=SSPSTAT & 0x2d;
   output_bit(PIN_A4,0);

   switch(s){
   case 0x9:    //estado 1
      s=i2c_read();
      index=0;
      break;

   case 0x21:
   case 0x29:   //estado 2
      muestras.b[index]=i2c_read();
      index++;
      if(index==32)
         index=0;
      break;

   case 0x0c:  //estado 3
      index=0;
      i2c_write(muestras.b[index]);
      index++;
      break;

   case 0x2c:  //estado 4
      if(index==32)
         index=0;
      i2c_write(muestras.b[index]);
      index++;
      break;

   case 0x20:
   case 0x28:  //estado 5 fin de lectura
      break;

   default:    // error (no deberia pasar)
      reset_cpu();
   }
}

void init_pic(){
   setup_adc_ports( NO_ANALOGS );
   setup_adc(ADC_OFF);
   output_a(0);
   set_tris_b(0xFF);
   ext_int_edge( L_TO_H );
   setup_timer_1(T1_INTERNAL|T1_DIV_BY_8);
   set_timer1(3036);

   enable_interrupts(INT_SSP);
   enable_interrupts(INT_EXT);
   enable_interrupts(INT_TIMER1);
}


void main(){
   int i,n;
   setup_wdt(WDT_144MS);
   init_pic();
   init_nrf();
   set_rx_mode();
   nuevo=0;
   id_paq=0;
   contador_t1=0;
   nsensores=0;
   sensores=0;
   dlistos=0;
   registro=0;
   i2c_estado=ESPERANDO;

   for(i=0;i<32;i=i+2){
      muestras.b[i]=200;
      muestras.b[i+1]=0xCE; // -50
   }

   enable_interrupts(GLOBAL);
   while(1){
      restart_wdt();
      if(nuevo){
         nuevo=0;
         output_bit(PIN_A4,1);
      }
   }
}
